import unittest

import torch

from normalkl import kl


class TestKL(unittest.TestCase):
    def setUp(self):
        self.dtype = torch.float64

    def test_kl_readme_example(self):
        dtype = self.dtype
        mean1 = torch.tensor([4.0, 5.0], dtype=dtype)
        covariance1 = torch.tensor([[1.0, 1.0], [2.0, 4.0]], dtype=dtype)
        mean2 = torch.tensor([1.0, 2.0], dtype=dtype)
        scalarvar2 = torch.tensor([3.0], dtype=dtype)

        result = kl(mean1, 'covmat', covariance1, mean2, 'scalarvar', scalarvar2)

        expected = torch.tensor([3.5853720317214703], dtype=dtype)
        self.assertTrue(torch.allclose(expected, result))


# Run the tests
if __name__ == '__main__':
    unittest.main(argv=[''], exit=False)




import unittest

import torch

from normalkl import kl, optimal_covariance


class TestOptimalCov2(unittest.TestCase):
    def setUp(self):
        dtype = torch.float64

        d1, d2 = 3, 4
        d = d1 * d2

        dtype=torch.float64

        # Means
        self.mean1 = torch.randn(d, dtype=dtype)
        self.mean2 = torch.randn(d, dtype=dtype)

        # Full covariance 
        self.L1 = torch.randn(d, d, dtype=dtype).tril()
        self.L2 = torch.randn(d, d, dtype=dtype).tril()
        self.covmat1 = self.L1 @ self.L1.T
        self.covmat2 = self.L2 @ self.L2.T
            
        # Cholesky factors for cholkroncov1 and cholkroncov2
        self.LU1 = torch.randn(d1, d1, dtype=dtype).tril()
        self.LV1 = torch.randn(d2, d2, dtype=dtype).tril()
        self.LU2 = torch.randn(d1, d1, dtype=dtype).tril()
        self.LV2 = torch.randn(d2, d2, dtype=dtype).tril()

        # Covariance matrices for cholkroncov1 and cholkroncov2
        self.U1 = self.LU1 @ self.LU1.T
        self.V1 = self.LV1 @ self.LV1.T
        self.U2 = self.LU2 @ self.LU2.T
        self.V2 = self.LV2 @ self.LV2.T

        # Cholesky Kronecker decompositions
        self.cholkroncov1 = (self.LU1, self.LV1)
        self.cholkroncov2 = (self.LU2, self.LV2)

        # Full covariance matrices
        self.kron_covmat1 = torch.kron(self.U1, self.V1)
        self.kron_covmat2 = torch.kron(self.U2, self.V2)

        # Other inputs
        self.I = torch.eye(d, dtype=dtype)
        self.v1 = torch.randn(d, dtype=dtype)
        self.v2 = torch.randn(d, dtype=dtype)
        self.s2 = torch.tensor(2.0, dtype=dtype)
        self.a1 = torch.randn(d1, dtype=dtype).square()
        self.b1 = torch.randn(d2, dtype=dtype).square()

        self.rtol = 1e-6
        self.atol = 1e-6

        self.cov1_list = [('covmat', self.covmat1),
                     ('precmat', self.covmat1),
                     ('diagvar', self.v1),
                     ('diagprec', self.v1),
                     ('scalarvar', self.s2),
                     ('scalarprec', self.s2),
                     ('identity', None),
                     ('cholcov', self.L1),
                     ('cholprec', self.L1),
                     ('kroncov', (self.U1, self.V1)),
                     ('kronprec', (self.U1, self.U2)),
                     ('cholkroncov', (self.LU1, self.LV1)),
                     ('cholkronprec', (self.LU1, self.LV1)),
                     ('diagcovrowcol', (self.a1, self.b1)),
                     ('diagcovrow', self.a1),
                     ('diagcovcol', self.b1)]



    def _is_optimum(self, objective_function, point, tolerance=1e-3):
        """
        Check if the point is at an optimum of the objective function.

        Args:
            point (torch.Tensor): The point to check.
            objective_function (callable): The objective function to minimize or maximize.
            tolerance (float): The tolerance for checking if the gradient is close to zero.

        Returns:
            bool: True if the point is at the optimum, False otherwise.
        """
        # Ensure the point requires gradients for optimization checking
        point = point.clone().detach().requires_grad_(True)

        # Compute the objective function value at the given point
        obj_value = objective_function(point)
        
        # Compute the gradients with respect to the point
        obj_value.backward()

        # Check if the gradients are close to zero within the specified tolerance
        is_at_optimum = torch.all(torch.abs(point.grad) < tolerance)

        return is_at_optimum


    # test_optimal_*_diagvar

    def test_optimal_covmat_diagvar(self):
        optimal_diagvar2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvar')
        kl_func = lambda diagvar2: kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvar', diagvar2)

        self.assertTrue(self._is_optimum(kl_func, optimal_diagvar2))

    # test_optimal_*_diagprec

    def test_optimal_covmat_diagprec(self):
        optimal_diagprec2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagprec')
        kl_func = lambda diagprec2: kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagprec', diagprec2)

        self.assertTrue(self._is_optimum(kl_func, optimal_diagprec2))

    # test_optimal_*_scalarvar

    def test_optimal_covmat_scalarvar(self):
        optimal_scalarvar2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarvar')
        kl_func = lambda scalarvar2: kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarvar', scalarvar2)

        self.assertTrue(self._is_optimum(kl_func, optimal_scalarvar2))

    def test_optimal_cholkroncov_scalarvar(self):
        optimal_scalarvar2 = optimal_covariance(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarvar')
        kl_func = lambda scalarvar2: kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarvar', scalarvar2)

        self.assertTrue(self._is_optimum(kl_func, optimal_scalarvar2))

    # test_optimal_*_scalarprec

    def test_optimal_covmat_scalarprec(self):
        optimal_scalarprec2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarprec')
        kl_func = lambda scalarprec2: kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarprec', scalarprec2)

        self.assertTrue(self._is_optimum(kl_func, optimal_scalarprec2))


    def test_optimal_cholkroncov_scalarprec(self):
        optimal_scalarprec2 = optimal_covariance(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarprec')
        kl_func = lambda scalarprec2: kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarprec', scalarprec2)

        self.assertTrue(self._is_optimum(kl_func, optimal_scalarprec2))




# Run the tests
if __name__ == '__main__':
    unittest.main(argv=[''], exit=False)




import torch

from .logdet import *

# auto_kl_*_diagvar
def auto_kl_covmat_diagvar(mean1, covmat1, mean2):
    """ Compute KL with automatic optimal diagonal variance according to
        argmin_v3 KL( N(v1, M1) || M(v2, diag(v3)) )
    """
    mean_diff = mean1 - mean2
    covmat1_diag = covmat1.diag()
    return 0.5 * (torch.sum(torch.log(covmat1_diag + mean_diff.square())) - logdet_covmat(covmat1))

def auto_kl_cholkroncov_diagvar(mean1, cholkroncov1, mean2):
    """ Compute KL with automatic optimal diagonal variance according to
        argmin_v3 KL( N(v1, M1) || M(v2, diag(v3)) )
    """
    mean_diff = mean1 - mean2
    LU1, LV1 = cholkroncov1
    covmat1_diag = torch.kron(LU1.square().sum(1), LV1.square().sum(1))

    return 0.5 * (torch.sum(torch.log(covmat1_diag + mean_diff.square())) - logdet_cholkroncov(cholkroncov1))

# auto_kl_*_diagprec
def auto_kl_covmat_diagprec(mean1, covmat1, mean2):
    return auto_kl_covmat_diagvar(mean1, covmat1, mean2)

# auto_kl_*_scalarvar
def auto_kl_covmat_scalarvar(mean1, covmat1, mean2):
    """ Compute KL with automatic optimal diagonal variance according to
        argmin_v3 KL( N(v1, M1) || M(v2, v3 I))
	"""
    d = mean1.size(0)
    mean_diff = mean1 - mean2
    total_variance = torch.trace(covmat1) + mean_diff.square().sum()
    return 0.5 * (d * torch.log(total_variance / d) - logdet_covmat(covmat1))

def auto_kl_cholkroncov_scalarvar(mean1, cholkroncov1, mean2):
    """ Compute KL with automatic optimal diagonal variance according to
        argmin_v3 KL( N(v1, M1) || M(v2, v3 I))
	"""
    d = mean1.size(0)
    LU, LV = cholkroncov1
    cov1_trace = LU.square().sum() * LV.square().sum()

    mean_diff = mean1 - mean2

    total_variance = cov1_trace + mean_diff.square().sum()
    return 0.5 * (d * torch.log(total_variance / d) - logdet_cholkroncov(cholkroncov1))

# auto_kl_*_scalarprec
def auto_kl_covmat_scalarprec(mean1, covmat1, mean2):
    return auto_kl_covmat_scalarvar(mean1, covmat1, mean2)

def auto_kl_cholkroncov_scalarprec(mean1, cholkroncov1, mean2):
    return auto_kl_cholkroncov_scalarvar(mean1, cholkroncov1, mean2)


def auto_kl(mean1, cov_type1, cov1, mean2, cov_type2):
    """ Compute KL with automatic optimal covariance according to
            argmin_cov2 KL( N(mean1, cov1) || N(mean2, cov2))
        and constrained by type of covariance (cov_type2)
    """
    func_name = f"auto_kl_{cov_type1}_{cov_type2}"

    func = globals()[func_name]

    mean_diff = mean1 - mean2

    return func(mean1, cov1, mean2)





""" Functions that compute quadratic term of the KL divergence for different covariance types. """

import torch


# Precision versions of the above functions:

def quadratic_covmat(v1, covmat2):
    """ Return quadratic term: v1^T covmat22^{-1} v1 / 2 """
    return 0.5 * (v1.unsqueeze(0) @ torch.inverse(covmat2) @ v1).squeeze()

def quadratic_precmat(v1, precmat2):
    """ Return quadratic term: v1^T precmat2 v1 / 2 """
    return 0.5 * (v1.unsqueeze(0) @ precmat2 @ v1).squeeze()

def quadratic_diagvar(v1, diagvar2):
    """ Return quadratic term: v1^T diag(var)^{-1} v1 / 2"""
    return 0.5 * torch.sum(v1.square() / diagvar2).squeeze()

def quadratic_diagprec(v1, diagprec2):
    """ Return quadratic term: v1^T diag(prec2) v1 / 2 """
    return 0.5 * torch.sum(v1.square() * diagprec2).squeeze()

def quadratic_scalarvar(v1, scalarvar2):
    """ Return quadratic term: v1^T (scalarvar2 * I)^{-1} v1 / 2 """
    return 0.5 * torch.sum(v1.square() / scalarvar2).squeeze()

def quadratic_scalarprec(v1, scalarprec2):
    """ Return quadratic term: v1^T (scalarprec * I) v1 / 2 """
    return 0.5 * torch.sum(v1.square() * scalarprec2).squeeze()

def quadratic_identity(v1, _ignore):
    """ Return quadratic term: v1^T v1 / 2 """
    return 0.5 * v1.square().sum().squeeze()

def quadratic_cholcov(v1, cholcov2):
    """ Return quadratic term: v1^T (L2 L2.T )^{-1} v1 / 2 
    = v1^T L2^{-1}^T L2^{-1} v1 / 2 
    = (L2^{-1} v1)^2 / 2
    """
    L_inv = torch.inverse(cholcov2) 
    return 0.5 * torch.sum((L_inv @ v1) ** 2)

def quadratic_cholprec(v1, cholprec2):
    """ Return quadratic term: v1^T (L2 L2^T) v1 / 2 
    = v1^T L2 (L2^T v1)
    = (L2^T v1)^T (L2^T v1)
    """
    LTv = cholprec2.T @ v1
    return 0.5 * torch.sum(LTv.square())

def quadratic_kroncov(v1, kroncov2):
    """ Return quadratic term: v1^T (U2 kron V2)^{-1} v1 / 2 """
    U2, V2 = kroncov2
    d1, d2 = len(U2), len(V2)
    v1_mat = v1.view(d1, d2)  # Reshape v1 to a matrix of shape (d1, d2)
    U2_inv = torch.inverse(U2)
    V2_inv = torch.inverse(V2)
    return 0.5 * torch.trace(U2_inv @ v1_mat @ V2_inv @ v1_mat.T)

def quadratic_kronprec(v1, kronprec2):
    """ Return quadratic term: v1^T (U2 kron V2) v1 / 2 """
    U2, V2 = kronprec2
    d1, d2 = len(U2), len(V2)
    v1_mat = v1.view(d1, d2)  # Reshape v1 to a matrix of shape (d1, d2)
    return 0.5 * torch.trace(U2 @ v1_mat @ V2 @ v1_mat.T)

def quadratic_cholkroncov(v1, cholkroncov2):
    """ Return quadratic term: v1^T (L_U2 L_U2.T kron L_V2 L_V2.T)^{-1} v1 / 2 """
    LU2, LV2 = cholkroncov2
    d1, d2 = len(LU2), len(LV2)
    v1_mat = v1.view(d1, d2)  # Reshape v1 to a matrix of shape (d1, d2)
    LU2_inv = torch.inverse(LU2)
    LV2_inv = torch.inverse(LV2)
    return 0.5 * v1.unsqueeze(0) @ torch.kron(LU2_inv.T @ LU2_inv, LV2_inv.T @ LV2_inv) @ v1

def quadratic_cholkronprec(v1, cholkronprec2):
    """ Return quadratic term: v1^T (L_U2 L_U2.T kron L_V2 L_V2.T) v1 / 2 """
    LU2_inv, LV2_inv = cholkronprec2
    d1, d2 = len(LU2_inv), len(LV2_inv)
    v1_mat = v1.view(d1, d2)  # Reshape v1 to a matrix of shape (d1, d2)
    return 0.5 * v1.unsqueeze(0) @ torch.kron(LU2_inv.T @ LU2_inv, LV2_inv.T @ LV2_inv) @ v1

def quadratic_diagcovrowcol(v1, a2, b2):
    """ Return quadratic term: v1^T diag(vec(b2 a2^T))^{-1} v1 / 2 """
    ab = torch.outer(a2, b2)
    ab_diag = ab.view(-1)
    return 0.5 * torch.sum(v1.square() / ab_diag)

def quadratic_diagcovrow(v1, a2):
    """ Return quadratic term: v1^T diag(vec(1 a2^T))^{-1} v1 / 2 """
    d1 = len(a2)
    d2 = len(v1) // d1
    return 0.5 * torch.sum(v1.view(d1, d2).square() / a2.unsqueeze(1))

def quadratic_diagcovcol(v1, b2):
    """ Return quadratic term: v1^T diag(vec(b2 1^T))^{-1} v1 / 2"""
    d2 = len(b2)
    d1 = len(v1) // d2
    return 0.5 * torch.sum(v1.view(d1, d2).square() / b2.view(1, -1))








import unittest

from normalkl.logdet import *

class TestLogDet(unittest.TestCase):
    def setUp(self):
        d1, d2 = 3, 4
        d = d1 * d2

        dtype = torch.float64

        self.LU1 = torch.randn(d1, d1, dtype=dtype).tril()
        self.LV1 = torch.randn(d2, d2, dtype=dtype).tril()
        self.U1 = self.LU1 @ self.LU1.T
        self.V1 = self.LV1 @ self.LV1.T
        self.kron_covmat1 = torch.kron(self.U1, self.V1)
        
        self.L1 = torch.randn(d, d, dtype=dtype).tril() + torch.eye(d, dtype=dtype)
        self.covmat1 = self.L1 @ self.L1.T
        
        self.I = torch.eye(d)
        self.s2 = torch.tensor(2.0)
        self.v1 = torch.rand(d)

    def test_logdet_covmat(self):
        expected = torch.logdet(self.covmat1)
        result = logdet_covmat(self.covmat1)
        self.assertTrue(torch.allclose(expected, result))

    def test_logdet_precmat(self):
        expected = -torch.logdet(self.covmat1)
        result = logdet_precmat(self.covmat1)
        self.assertTrue(torch.allclose(expected, result))

    def test_logdet_diagvar(self):
        expected = self.v1.abs().log().sum()
        result = logdet_diagvar(self.v1)
        self.assertTrue(torch.allclose(expected, result))

    def test_logdet_diagprec(self):
        expected = -self.v1.abs().log().sum()
        result = logdet_diagprec(self.v1)
        self.assertTrue(torch.allclose(expected, result))

    def test_logdet_scalarvar(self):
        dim = self.covmat1.size(0)
        expected = dim * self.s2.abs().log()
        result = logdet_scalarvar(self.s2, dim)
        self.assertTrue(torch.allclose(expected, result))

    def test_logdet_scalarprec(self):
        dim = self.covmat1.size(0)
        expected = -dim * self.s2.log()
        result = logdet_scalarprec(self.s2, dim)
        self.assertTrue(torch.allclose(expected, result))

    def test_logdet_cholcov(self):
        expected = self.L1.diag().square().log().sum()
        result = logdet_cholcov(self.L1)
        self.assertTrue(torch.allclose(expected, result))

    def test_logdet_cholprec(self):
        expected = -self.L1.diag().square().log().sum()
        result = logdet_cholprec(self.L1)
        self.assertTrue(torch.allclose(expected, result))

    def test_logdet_kroncov(self):
        expected = len(self.V1) * logdet_covmat(self.U1) + len(self.U1) * logdet_covmat(self.V1)
        result = logdet_kroncov((self.U1, self.V1))
        self.assertTrue(torch.allclose(expected, result))

    def test_logdet_kronprec(self):
        expected = len(self.V1) * logdet_precmat(self.U1) + len(self.U1) * logdet_precmat(self.V1)
        result = logdet_kronprec((self.U1, self.V1))
        self.assertTrue(torch.allclose(expected, result))

    def test_logdet_cholkroncov(self):
        expected = len(self.LV1) * logdet_cholcov(self.LU1) + len(self.LU1) * logdet_cholcov(self.LV1)
        result = logdet_cholkroncov((self.LU1, self.LV1))
        self.assertTrue(torch.allclose(expected, result))

    def test_logdet_cholkronprec(self):
        expected = len(self.LV1) * logdet_cholprec(self.LU1) + len(self.LU1) * logdet_cholprec(self.LV1)
        result = logdet_cholkronprec((self.LU1, self.LV1))
        self.assertTrue(torch.allclose(expected, result))

# Run the tests
if __name__ == '__main__':
    unittest.main(argv=[''], exit=False)




import unittest
import torch

from normalkl.quadratic import *

class TestQuadratic(unittest.TestCase):
    def setUp(self):
        torch.manual_seed(0)

        d1, d2 = 3, 4
        d = d1 * d2

        dtype = torch.float64

        # Full covariance
        self.L1 = torch.randn(d, d, dtype=dtype).tril()
        self.L2 = torch.randn(d, d, dtype=dtype).tril()
        self.L1_inv = torch.inverse(self.L1)
        self.L2_inv = torch.inverse(self.L2)
        self.covmat1 = self.L1 @ self.L1.T
        self.covmat2 = self.L2 @ self.L2.T

        # Cholesky factors for cholkroncov1 and cholkroncov2
        self.LU1 = torch.randn(d1, d1, dtype=dtype).tril()
        self.LV1 = torch.randn(d2, d2, dtype=dtype).tril()
        self.LU2 = torch.randn(d1, d1, dtype=dtype).tril()
        self.LV2 = torch.randn(d2, d2, dtype=dtype).tril()
        self.LU1_inv = torch.inverse(self.LU1)
        self.LV1_inv = torch.inverse(self.LV1)
        self.LU2_inv = torch.inverse(self.LU2)
        self.LV2_inv = torch.inverse(self.LV2)

        # Covariance matrices for cholkroncov1 and cholkroncov2
        self.U1 = self.LU1 @ self.LU1.T
        self.V1 = self.LV1 @ self.LV1.T
        self.U2 = self.LU2 @ self.LU2.T
        self.V2 = self.LV2 @ self.LV2.T
        self.U1_inv = torch.inverse(self.U1)
        self.V1_inv = torch.inverse(self.V1)
        self.U2_inv = torch.inverse(self.U2)
        self.V2_inv = torch.inverse(self.V2)

        # Cholesky Kronecker decompositions
        self.cholkroncov1 = (self.LU1, self.LV1)
        self.cholkroncov2 = (self.LU2, self.LV2)
        self.cholkronprec1 = (self.LU1_inv, self.LV1_inv)
        self.cholkronprec2 = (self.LU2_inv, self.LV2_inv)

        # Full covariance matrices
        self.kron_covmat1 = torch.kron(self.U1, self.V1)
        self.kron_covmat2 = torch.kron(self.U2, self.V2)
        self.kron_precmat1 = torch.kron(self.U1_inv, self.V1_inv)
        self.kron_precmat2 = torch.kron(self.U2_inv, self.V2_inv)

        # Other inputs
        self.I = torch.eye(d, dtype=dtype)
        self.v1 = torch.randn(d, dtype=dtype)
        self.v2 = torch.randn(d, dtype=dtype)
        self.s2 = torch.tensor(2.0, dtype=dtype)
        self.a2 = torch.randn(d1, dtype=dtype).square()
        self.b2 = torch.randn(d2, dtype=dtype).square()

        # Diagonal covariance and precision
        self.diagvar = torch.diag(self.covmat1)
        self.diagprec = 1.0 / self.diagvar

        # Scalar variance and precision
        self.scalarvar = torch.mean(self.diagvar).unsqueeze(0)
        self.scalarprec = 1.0 / self.scalarvar

        # Precision matrix
        self.precmat1 = torch.inverse(self.covmat1)
        self.precmat2 = torch.inverse(self.covmat2)

        # Tolerances
        self.rtol = 1e-6
        self.atol = 1e-6

    def test_quadratic_covmat(self):
        result = quadratic_covmat(self.v1, self.covmat1)
        expected = quadratic_covmat(self.v1, self.covmat1)
        self.assertTrue(torch.isclose(result, expected, rtol=self.rtol, atol=self.atol))

    def test_quadratic_precmat(self):
        result = quadratic_precmat(self.v1, self.precmat1)
        expected = quadratic_covmat(self.v1, self.covmat1)
        self.assertTrue(torch.isclose(result, expected, rtol=self.rtol, atol=self.atol))

    def test_quadratic_diagvar(self):
        result = quadratic_diagvar(self.v1, self.diagvar)
        expected = quadratic_covmat(self.v1, torch.diag(self.diagvar))
        self.assertTrue(torch.isclose(result, expected, rtol=self.rtol, atol=self.atol))

    def test_quadratic_diagprec(self):
        result = quadratic_diagprec(self.v1, self.diagprec)
        expected = quadratic_covmat(self.v1, torch.diag(1.0 / self.diagprec))
        self.assertTrue(torch.isclose(result, expected, rtol=self.rtol, atol=self.atol))

    def test_quadratic_scalarvar(self):
        result = quadratic_scalarvar(self.v1, self.scalarvar)
        expected = quadratic_covmat(self.v1, self.scalarvar * torch.eye(self.v1.shape[0], dtype=self.v1.dtype))
        self.assertTrue(torch.isclose(result, expected, rtol=self.rtol, atol=self.atol))

    def test_quadratic_scalarprec(self):
        result = quadratic_scalarprec(self.v1, self.scalarprec)
        expected = quadratic_covmat(self.v1, torch.eye(self.v1.shape[0], dtype=self.v1.dtype) / self.scalarprec.item())
        self.assertTrue(torch.isclose(result, expected, rtol=self.rtol, atol=self.atol))

    def test_quadratic_identity(self):
        result = quadratic_identity(self.v1, self.I)
        expected = quadratic_covmat(self.v1, torch.eye(self.v1.shape[0], dtype=self.v1.dtype))
        self.assertTrue(torch.isclose(result, expected, rtol=self.rtol, atol=self.atol))

    def test_quadratic_cholcov(self):
        result = quadratic_cholcov(self.v1, self.L2)
        expected = quadratic_covmat(self.v1, self.covmat2)
        self.assertTrue(torch.isclose(result, expected, rtol=self.rtol, atol=self.atol))

    def test_quadratic_cholprec(self):
        result = quadratic_cholprec(self.v1, self.L2)
        expected = quadratic_covmat(self.v1, self.precmat2)
        self.assertTrue(torch.isclose(result, expected, rtol=self.rtol, atol=self.atol))

    def test_quadratic_kroncov(self):
        result = quadratic_kroncov(self.v1, (self.U2, self.V2))
        expected = quadratic_covmat(self.v1, self.kron_covmat2)
        self.assertTrue(torch.isclose(result, expected, rtol=self.rtol, atol=self.atol))

    def test_quadratic_kronprec(self):
        result = quadratic_kronprec(self.v1, (self.U2_inv, self.V2_inv))
        expected = quadratic_covmat(self.v1, self.kron_covmat2)
        self.assertTrue(torch.isclose(result, expected, rtol=self.rtol, atol=self.atol))

    def test_quadratic_cholkroncov(self):
        result = quadratic_cholkroncov(self.v1, (self.LU2, self.LV2))
        expected = quadratic_covmat(self.v1, self.kron_covmat2)
        print(result, expected)
        self.assertTrue(torch.isclose(result, expected, rtol=self.rtol, atol=self.atol))

    def test_quadratic_cholkronprec(self):
        result = quadratic_cholkronprec(self.v1, (self.LU2_inv, self.LV2_inv))
        expected = quadratic_covmat(self.v1, self.kron_covmat2)
        self.assertTrue(torch.isclose(result, expected, rtol=self.rtol, atol=self.atol))





# Run the tests
if __name__ == '__main__':
    unittest.main(argv=[''], exit=False)




""" Functions that compute log-determinant term of the KL divergence for different covariance types. """

import torch

def logdet_covmat(covmat):
    """ Log determinant for full covariance matrix """
    return torch.logdet(covmat)

def logdet_precmat(precmat):
    """ Log determinant for full precision matrix """
    return -torch.logdet(precmat)

def logdet_diagvar(diagvar):
    """ Log determinant for diagonal covariance matrix (vector) """
    return diagvar.abs().log().sum().squeeze()

def logdet_scalarvar(scalarvar, dim):
    """ Log determinant for scalar variance """
    return dim * scalarvar.abs().log().squeeze()

def logdet_cholcov(cholcov):
    """ Log determinant for Cholesky covariance matrix """
    L = cholcov
    return L.diag().square().log().sum()

def logdet_kroncov(kroncov):
    """ Log determinant for Kronecker product of two matrices """
    U, V = kroncov
    return len(V) * logdet_covmat(U) + len(U) * logdet_covmat(V)

def logdet_cholkroncov(cholkroncov):
    LU, LV = cholkroncov
    """ Log determinant for Cholesky Kronecker covariance """
    return len(LV) * logdet_cholcov(LU) + len(LU) * logdet_cholcov(LV)


# Precision versions of the above functions:

def logdet_precmat(precmat):
    """ Log determinant for precision matrix """
    return -logdet_covmat(precmat)

def logdet_diagprec(diagprec):
    """ Log determinant for diagonal precision matrix """
    return -diagprec.abs().log().sum()

def logdet_scalarprec(scalarprec, dim):
    """ Log determinant for scalar precision """
    return -dim * scalarprec.log()

def logdet_cholprec(cholprec):
    """ Log determinant for Cholesky precision matrix """
    L = cholprec
    return -L.diag().square().log().sum()

def logdet_kronprec(kronprec):
    U, V = kronprec
    """ Log determinant for Kronecker product of two precision matrices """
    return len(V) * logdet_precmat(U) + len(U) * logdet_precmat(V)

def logdet_cholkronprec(cholkronprec):
    """ Log determinant for Cholesky Kronecker precision """
    LU, LV = cholkronprec
    return len(LV) * logdet_cholprec(LU) + len(LU) * logdet_cholprec(LV)




import torch

# optimal_*_diagvar
def optimal_covmat_diagvar(mean_diff, covmat1):
    """ Compute optimal diagonal variance of
        argmin_v3 KL( N(v1, M1) || M(v2, diag(v3)) )
    """
    return covmat1.diag() + mean_diff.square()

def optimal_cholkroncov_diagvar(mean_diff, cholkroncov1):
    """ Compute optimal diagonal variance of
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T)) || M(v2, diag(v3)) )
    """
    LU1, LV1 = cholkroncov1
    covmat1_diag = torch.kron(LU1.square().sum(1), LV1.square().sum(1))
    return covmat1_diag + mean_diff.square()

# optimal_*_diagprec
def optimal_covmat_diagprec(mean_diff, covmat1):
    """ Compute optimal diagonal precision of
        argmin_v3 KL( N(v1, M1) || M(v2, diag(v3)^{-1}) )
    """
    return 1 / (covmat1.diag() + mean_diff.square())

def optimal_cholkroncov_diagprec(mean_diff, cholkroncov1):
    """ Compute optimal diagonal variance of
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T)) || M(v2, diag(v3)) )
    """
    LU1, LV1 = cholkrncov1
    covmat1_diag = torch.kron(LU1.square().sum(1), LV1.square().sum(1))
    return 1 / (covmat1_diag + mean_diff.square())

# optimal_*_scalarvar
def optimal_covmat_scalarvar(mean_diff, covmat1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, covmat1) || M(v2, v3 I))
    """
    return (torch.trace(covmat1) + mean_diff.square().sum()) / len(mean_diff)

def optimal_precmat_scalarvar(mean_diff, precmat1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, precmat1^{-1}) || M(v2, v3 I))
    """
    return (torch.trace(torch.inverse(precmat1)) + mean_diff.square().sum()) / len(mean_diff)

def optimal_diagvar_scalarvar(mean_diff, diagvar1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, diagvar1) || M(v2, v3 I))
    """
    return (torch.sum(diagvar1) + mean_diff.square().sum()) / len(mean_diff)

def optimal_diagprec_scalarvar(mean_diff, diagprec1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, diagprec1) || M(v2, v3 I))
    """
    return (torch.sum(1/diagprec1) + mean_diff.square().sum()) / len(mean_diff)

def optimal_scalarvar_scalarvar(mean_diff, scalarvar1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, scalarvar1 I ) || M(v2, v3 I))
    """
    d = mean_diff.size(0)
    return (d * scalarvar1 + mean_diff.square().sum()) / len(mean_diff)

def optimal_scalarprec_scalarvar(mean_diff, scalarprec1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, scalarprec1^{-1} I ) || M(v2, v3 I))
    """
    d = mean_diff.size(0)
    return (d * scalarprec1 + mean_diff.square().sum()) / len(mean_diff)

def optimal_identity_scalarvar(mean_diff, _ignore):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, I ) || M(v2, v3 I))
    """
    d = mean_diff.size(0)
    return (mean_diff.square().sum()) / len(mean_diff)

def optimal_cholcov_scalarvar(mean_diff, cholcov1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, L1 L1^T ) || M(v2, v3 I))
    """
    trace1 = cholcov.square().sum()
    return (trace1 + mean_diff.square().sum()) / len(mean_diff)

def optimal_cholprec_scalarvar(mean_diff, cholcov1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, L1 L1^T ) || M(v2, v3 I))
    """
    trace1 = torch.inverse(cholcov1).square().sum()
    return (trace1 + mean_diff.square().sum()) / len(mean_diff)

def optimal_kroncov_scalarvar(mean_diff, kroncov1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, (U1 kron V1) ) || M(v2, v3 I))
    """
    U1, V1 = kroncov1
    trace1 = torch.trace(U1) * torch.trace(V1)
    return (trace1 + mean_diff.square().sum()) / len(mean_diff)

def optimal_kronprec_scalarvar(mean_diff, kronprec1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, (U1 kron V1) ) || M(v2, v3 I))
    """
    U1, V1 = kronprec1
    trace1 = torch.trace(torch.inverse(U1)) * torch.trace(torch.inverse(V1))
    return (trace1 + mean_diff.square().sum()) / len(mean_diff)

def optimal_cholkroncov_scalarvar(mean_diff, cholkroncov1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T) ) || M(v2, v3 I))
    """
    LU1, LV1 = cholkroncov1
    trace1 = LU1.square().sum() * LV1.square().sum()
    return (trace1 + mean_diff.square().sum()) / len(mean_diff)

def optimal_cholkronprec_scalarvar(mean_diff, cholkronprec1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T) ) || M(v2, v3 I))
    """
    LU1_inv, LV1_inv = cholkronprec1
    LU1, LV1 = torch.inverse(LU1_inv), torch.inverse(LV1_inv)
    trace1 = LU1.square().sum() * LV1.square().sum()
    return (trace1 + mean_diff.square().sum()) / len(mean_diff)

def optimal_diagcovrowcol_scalarvar(mean_diff, diagcovrowcol1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, diag(a) kron diag(b) ) || M(v2, v3 I))
    """
    a, b = diagcovrowcol1
    trace1 = torch.outer(a, b).sum()
    return (trace1 + mean_diff.square().sum()) / len(mean_diff)

def optimal_diagcovrow_scalarvar(mean_diff, diagcovrow1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, diag(a) kron diag(1) ) || M(v2, v3 I))
    """
    d = mean_diff.size(0)
    d1 = diagcovrow1.size(0)
    d2 = d // d1
    trace1 = diagcovrow1.sum() * d2
    return (trace1 + mean_diff.square().sum()) / len(mean_diff)

def optimal_diagcovcol_scalarvar(mean_diff, diagcovcol1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, diag(1) kron diag(b) ) || M(v2, v3 I))
    """
    d = mean_diff.size(0)
    d2 = diagcovcol1.size(0)
    d1 = d // d2
    trace1 = d1 * diagcovcol1.sum()
    return (trace1 + mean_diff.square().sum()) / len(mean_diff)


# optimal_*_scalarprec
def optimal_covmat_scalarprec(mean_diff, covmat1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, covmat1) || M(v2, v3 I))
    """
    return len(mean_diff) / (torch.trace(covmat1) + mean_diff.square().sum()) 

def optimal_precmat_scalarprec(mean_diff, precmat1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, precmat1^{-1}) || M(v2, v3 I))
    """
    return len(mean_diff) / (torch.trace(torch.inverse(precmat1)) + mean_diff.square().sum()) 

def optimal_diagvar_scalarprec(mean_diff, diagvar1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, diagvar1) || M(v2, v3 I))
    """
    return len(mean_diff) / (torch.sum(diagvar1) + mean_diff.square().sum()) 

def optimal_diagprec_scalarprec(mean_diff, diagprec1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, diagprec1) || M(v2, v3 I))
    """
    return len(mean_diff) / (torch.sum(1/diagprec1) + mean_diff.square().sum()) 

def optimal_scalarprec_scalarprec(mean_diff, scalarvar1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, scalarvar1 I ) || M(v2, v3 I))
    """
    d = mean_diff.size(0)
    return len(mean_diff) / (d * scalarvar1 + mean_diff.square().sum()) 

def optimal_scalarprec_scalarprec(mean_diff, scalarprec1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, scalarprec1^{-1} I ) || M(v2, v3 I))
    """
    d = mean_diff.size(0)
    return len(mean_diff) / (d * scalarprec1 + mean_diff.square().sum()) 

def optimal_identity_scalarprec(mean_diff, _ignore):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, I ) || M(v2, v3 I))
    """
    d = mean_diff.size(0)
    return len(mean_diff) / (mean_diff.square().sum()) 

def optimal_cholcov_scalarprec(mean_diff, cholcov1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, L1 L1^T ) || M(v2, v3 I))
    """
    trace1 = cholcov.square().sum(1)
    return len(mean_diff) / (trace1 + mean_diff.square().sum()) 

def optimal_cholprec_scalarprec(mean_diff, cholcov1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, L1 L1^T ) || M(v2, v3 I))
    """
    trace1 = torch.inverse(cholcov1).square().sum(1)
    return len(mean_diff) / (trace1 + mean_diff.square().sum()) 

def optimal_kroncov_scalarprec(mean_diff, kroncov1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, (U1 kron V1) ) || M(v2, v3 I))
    """
    U1, V1 = kroncov1
    trace1 = torch.trace(U1) * torch.trace(V1)
    return len(mean_diff) / (trace1 + mean_diff.square().sum()) 

def optimal_kronprec_scalarprec(mean_diff, kronprec1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, (U1 kron V1) ) || M(v2, v3 I))
    """
    U1, V1 = kronprec1
    trace1 = torch.trace(torch.inverse(U1)) * torch.trace(torch.inverse(V1))
    return len(mean_diff) / (trace1 + mean_diff.square().sum()) 

def optimal_cholkroncov_scalarprec(mean_diff, cholkroncov1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T) ) || M(v2, v3 I))
    """
    LU1, LV1 = cholkroncov1
    trace1 = LU1.square().sum() * LV1.square().sum()
    return len(mean_diff) / (trace1 + mean_diff.square().sum()) 

def optimal_cholkronprec_scalarprec(mean_diff, cholkronprec1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T) ) || M(v2, v3 I))
    """
    LU1_inv, LV1_inv = cholkronprec1
    LU1, LV1 = torch.inverse(LU1_inv), torch.inverse(LV1_inv)
    trace1 = LU1.square().sum() * LV1.square().sum()
    return len(mean_diff) / (trace1 + mean_diff.square().sum()) 

def optimal_diagcovrowcol_scalarprec(mean_diff, diagcovrowcol1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, diag(a) kron diag(b) ) || M(v2, v3 I))
    """
    a, b = diagcovrowcol1
    trace1 = torch.outer(a, b).sum()
    return len(mean_diff) / (trace1 + mean_diff.square().sum()) 

def optimal_diagcovrow_scalarprec(mean_diff, diagcovrow1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, diag(a) kron diag(1) ) || M(v2, v3 I))
    """
    d = mean_diff.size(0)
    d1 = diagcovrow1.size(0)
    d2 = d // d1
    trace1 = diagcovrow1.sum() * d2
    return len(mean_diff) / (trace1 + mean_diff.square().sum())

def optimal_diagcovcol_scalarprec(mean_diff, diagcovcol1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, diag(1) kron diag(b) ) || M(v2, v3 I))
    """
    d = mean_diff.size(0)
    d2 = diagcovcol1.size(0)
    d1 = d // d2
    trace1 = d1 * diagcovcol1.sum()
    return len(mean_diff) / (trace1 + mean_diff.square().sum()) 



def optimal_covariance(mean1, cov_type1, cov1, mean2, cov_type2):
    """ Find optimal cov2 that solves
            argmin_cov2 KL( N(mean1, cov1) || N(mean2, cov2))
    """
    func_name = f"optimal_{cov_type1}_{cov_type2}"

    func = globals()[func_name]

    mean_diff = mean1 - mean2

    return func(mean_diff, cov1)





import torch
import unittest

from normalkl.trace import *

class TestTrace(unittest.TestCase):
    def setUp(self):
        torch.manual_seed(0)
        
        d1, d2 = 3, 4
        d = d1 * d2

        dtype=torch.float64

        # Full covariance 
        self.L1 = torch.randn(d, d, dtype=dtype).tril()
        self.L2 = torch.randn(d, d, dtype=dtype).tril()
        self.covmat1 = self.L1 @ self.L1.T
        self.covmat2 = self.L2 @ self.L2.T
            
        # Cholesky factors for cholkroncov1 and cholkroncov2
        self.LU1 = torch.randn(d1, d1, dtype=dtype).tril()
        self.LV1 = torch.randn(d2, d2, dtype=dtype).tril()
        self.LU2 = torch.randn(d1, d1, dtype=dtype).tril()
        self.LV2 = torch.randn(d2, d2, dtype=dtype).tril()

        # Covariance matrices for cholkroncov1 and cholkroncov2
        self.U1 = self.LU1 @ self.LU1.T
        self.V1 = self.LV1 @ self.LV1.T
        self.U2 = self.LU2 @ self.LU2.T
        self.V2 = self.LV2 @ self.LV2.T

        # Cholesky Kronecker decompositions
        self.cholkroncov1 = (self.LU1, self.LV1)
        self.cholkroncov2 = (self.LU2, self.LV2)

        # Full covariance matrices
        self.kron_covmat1 = torch.kron(self.U1, self.V1)
        self.kron_covmat2 = torch.kron(self.U2, self.V2)

        # Other inputs
        self.I = torch.eye(d, dtype=dtype)
        self.v1 = torch.randn(d, dtype=dtype)
        self.v2 = torch.randn(d, dtype=dtype)
        self.s2 = torch.tensor(2.0, dtype=dtype)
        self.a2 = torch.randn(d1, dtype=dtype).square()
        self.b2 = torch.randn(d2, dtype=dtype).square()

        self.rtol = 1e-6
        self.atol = 1e-6

    def test_trace_covmat_covmat(self):
        expected = trace_covmat_covmat(self.covmat1, self.covmat2)
        result = trace_covmat_covmat(self.covmat1, self.covmat2)
        self.assertTrue(torch.allclose(expected, result, rtol=self.rtol, atol=self.atol))

    def test_trace_covmat_precmat(self):       
        expected = trace_covmat_covmat(self.covmat1, torch.inverse(self.covmat2))
        result = trace_covmat_precmat(self.covmat1, self.covmat2)
        self.assertTrue(torch.allclose(expected, result, rtol=self.rtol, atol=self.atol))

    def test_trace_covmat_diagvar(self):
        expected = trace_covmat_covmat(self.covmat1, self.v2.diag())
        result = trace_covmat_diagvar(self.covmat1, self.v2)
        self.assertTrue(torch.allclose(expected, result, rtol=self.rtol, atol=self.atol))

    def test_trace_covmat_diagprec(self):
        expected = trace_covmat_covmat(self.covmat1, (1 / self.v2).diag())
        result = trace_covmat_diagprec(self.covmat1, self.v2)
        self.assertTrue(torch.allclose(expected, result, rtol=self.rtol, atol=self.atol))

    def test_trace_covmat_scalarvar(self):
        expected = trace_covmat_covmat(self.covmat1, self.s2 * self.I)
        result = trace_covmat_scalarvar(self.covmat1, self.s2)
        self.assertTrue(torch.allclose(expected, result, rtol=self.rtol, atol=self.atol))

    def test_trace_covmat_scalarprec(self):
        expected = trace_covmat_covmat(self.covmat1, (1 / self.s2) * self.I)
        result = trace_covmat_scalarprec(self.covmat1, self.s2)
        self.assertTrue(torch.allclose(expected, result, rtol=self.rtol, atol=self.atol))

    def test_trace_covmat_identity(self):
        expected = trace_covmat_covmat(self.covmat1, self.I)
        result = trace_covmat_identity(self.covmat1, _=None)
        self.assertTrue(torch.allclose(expected, result, rtol=self.rtol, atol=self.atol))
        
    def test_trace_covmat_cholcov(self):
        expected = trace_covmat_covmat(self.covmat1, self.kron_covmat2)
        result = trace_covmat_cholcov(self.covmat1, (self.U2, self.V2))
        print(self.U2 - torch.inverse(torch.inverse(self.U2)))
        print(self.V2 - torch.inverse(torch.inverse(self.V2)))
        # AA = torch.inverse(torch.kron(self.U2, self.V2))
        # BB = torch.kron(torch.inverse(self.U2), torch.inverse(self.V2))
        # print(AA - BB)
        print("EXP:", torch.trace(torch.inverse(torch.kron(self.U2, self.V2)) @ self.covmat1) / 2)
        print("EXP:", torch.trace(torch.kron(torch.inverse(self.U2), torch.inverse(self.V2)) @ self.covmat1) / 2)
        print(123, expected, result)
        self.assertTrue(torch.allclose(expected, result, rtol=self.rtol, atol=self.atol))

    def test_trace_covmat_cholprec(self):
        expected = trace_covmat_covmat(self.covmat1, torch.inverse(self.kron_covmat2))
        result = trace_covmat_cholprec(self.covmat1, (self.U2, self.V2))
        self.assertTrue(torch.allclose(expected, result, rtol=self.rtol, atol=self.atol))

    def test_trace_covmat_kroncov(self):
        expected = trace_covmat_covmat(self.covmat1, self.kron_covmat2)
        result = trace_covmat_kroncov(self.covmat1, (self.U2, self.V2))
        self.assertTrue(torch.allclose(expected, result, rtol=self.rtol, atol=self.atol))

    def test_trace_covmat_kronprec(self):
        expected = trace_covmat_covmat(self.covmat1, torch.inverse(self.kron_covmat2))
        result = trace_covmat_kronprec(self.covmat1, (self.U2, self.V2))
        self.assertTrue(torch.allclose(expected, result, rtol=self.rtol, atol=self.atol))

    def test_trace_covmat_cholkroncov(self):
        expected = trace_covmat_covmat(self.covmat1, self.kron_covmat2)
        result = trace_covmat_cholkroncov(self.covmat1, (self.LU2, self.LV2))
        self.assertTrue(torch.allclose(expected, result, rtol=self.rtol, atol=self.atol))

    def test_trace_covmat_cholkronprec(self):
        expected = trace_covmat_covmat(self.covmat1, torch.inverse(self.kron_covmat2))
        result = trace_covmat_cholkronprec(self.covmat1, (self.LU2, self.LV2))
        self.assertTrue(torch.allclose(expected, result, rtol=self.rtol, atol=self.atol))

    def test_trace_covmat_diagvarrowcol(self):
        expected = trace_covmat_covmat(self.covmat1, torch.outer(self.a2, self.b2).view(-1).diag())
        result = trace_covmat_diagvarrowcol(self.covmat1, (self.a2, self.b2))
        self.assertTrue(torch.allclose(expected, result, rtol=self.rtol, atol=self.atol))

    def test_trace_covmat_diagvarrow(self):
        d1 = len(self.a2)
        d2 = self.covmat1.size(0) // d1
        expected = trace_covmat_covmat(self.covmat1, self.a2.view(-1, 1).repeat(1, d2).view(-1).diag())
        result = trace_covmat_diagvarrow(self.covmat1, self.a2)
        self.assertTrue(torch.allclose(expected, result, rtol=self.rtol, atol=self.atol))

    def test_trace_covmat_diagvarcol(self):
        d2 = len(self.b2)
        d1 = self.covmat1.size(0) // d2
        expected = trace_covmat_covmat(self.covmat1, self.b2.view(1, -1).repeat(d1, 1).view(-1).diag())
        result = trace_covmat_diagvarcol(self.covmat1, self.b2)
        self.assertTrue(torch.allclose(expected, result, rtol=self.rtol, atol=self.atol))

    def test_trace_cholkroncov_covmat(self):
        result = trace_cholkroncov_covmat(self.cholkroncov1, self.covmat2)
        expected = trace_covmat_covmat(self.kron_covmat1, self.covmat2)
        self.assertTrue(torch.allclose(result, expected))

    def test_trace_cholkroncov_precmat(self):
        result = trace_cholkroncov_precmat(self.cholkroncov1, self.covmat2)
        expected = trace_covmat_covmat(self.kron_covmat1, torch.inverse(self.covmat2))
        self.assertTrue(torch.allclose(result, expected))

    def test_trace_cholkroncov_diagvar(self):
        result = trace_cholkroncov_diagvar(self.cholkroncov1, self.v2)
        expected = trace_covmat_covmat(self.kron_covmat1, torch.diag(self.v2))
        self.assertTrue(torch.allclose(result, expected))

    def test_trace_cholkroncov_diagprec(self):
        result = trace_cholkroncov_diagprec(self.cholkroncov1, self.v2)
        expected = trace_covmat_covmat(self.kron_covmat1, torch.diag(1 / self.v2))
        self.assertTrue(torch.allclose(result, expected))

    def test_trace_cholkroncov_scalarvar(self):
        result = trace_cholkroncov_scalarvar(self.cholkroncov1, self.s2)
        expected = trace_covmat_covmat(self.kron_covmat1, self.s2 * self.I)
        self.assertTrue(torch.allclose(result, expected))

    def test_trace_cholkroncov_scalarprec(self):
        result = trace_cholkroncov_scalarprec(self.cholkroncov1, self.s2)
        expected = trace_covmat_covmat(self.kron_covmat1, (1 / self.s2) * self.I)
        self.assertTrue(torch.allclose(result, expected))

    def test_trace_cholkroncov_identity(self):
        result = trace_cholkroncov_identity(self.cholkroncov1, None)
        expected = trace_covmat_covmat(self.kron_covmat1, self.I)
        self.assertTrue(torch.allclose(result, expected))

    def test_trace_cholkroncov_cholcov(self):
        result = trace_cholkroncov_cholcov(self.cholkroncov1, (self.U2, self.V2))
        expected = trace_covmat_covmat(self.kron_covmat1, self.kron_covmat2)
        self.assertTrue(torch.allclose(result, expected))

    def test_trace_cholkroncov_cholprec(self):
        result = trace_cholkroncov_cholprec(self.cholkroncov1, (self.U2, self.V2))
        expected = trace_covmat_covmat(self.kron_covmat1, torch.kron(torch.inverse(self.U2), torch.inverse(self.V2)))
        self.assertTrue(torch.allclose(result, expected))

    def test_trace_cholkroncov_kroncov(self):
        result = trace_cholkroncov_kroncov(self.cholkroncov1, (self.U2, self.V2))
        expected = trace_covmat_covmat(self.kron_covmat1, self.kron_covmat2)
        self.assertTrue(torch.allclose(result, expected))

    def test_trace_cholkroncov_kronprec(self):
        result = trace_cholkroncov_kronprec(self.cholkroncov1, (self.U2, self.V2))
        expected = trace_covmat_covmat(self.kron_covmat1, torch.inverse(self.kron_covmat2))
        self.assertTrue(torch.allclose(result, expected))

    def test_trace_cholkroncov_cholkroncov(self):
        result = trace_cholkroncov_cholkroncov(self.cholkroncov1, self.cholkroncov2)
        expected = trace_covmat_covmat(self.kron_covmat1, self.kron_covmat2)
        self.assertTrue(torch.allclose(result, expected))

    def test_trace_cholkroncov_cholkronprec(self):
        result = trace_cholkroncov_cholkronprec(self.cholkroncov1, self.cholkroncov2)
        expected = trace_covmat_covmat(self.kron_covmat1, torch.inverse(self.kron_covmat2))
        self.assertTrue(torch.allclose(result, expected))




import unittest

import torch

from normalkl import kl, auto_kl
from normalkl import optimal_covariance


class TestAutoKL(unittest.TestCase):
    def setUp(self):
        dtype = torch.float64

        d1, d2 = 3, 4
        d = d1 * d2

        dtype=torch.float64

        # Means
        self.mean1 = torch.randn(d, dtype=dtype)
        self.mean2 = torch.randn(d, dtype=dtype)

        # Full covariance 
        self.L1 = torch.randn(d, d, dtype=dtype).tril()
        self.L2 = torch.randn(d, d, dtype=dtype).tril()
        self.covmat1 = self.L1 @ self.L1.T
        self.covmat2 = self.L2 @ self.L2.T
            
        # Cholesky factors for cholkroncov1 and cholkroncov2
        self.LU1 = torch.randn(d1, d1, dtype=dtype).tril()
        self.LV1 = torch.randn(d2, d2, dtype=dtype).tril()
        self.LU2 = torch.randn(d1, d1, dtype=dtype).tril()
        self.LV2 = torch.randn(d2, d2, dtype=dtype).tril()

        # Covariance matrices for cholkroncov1 and cholkroncov2
        self.U1 = self.LU1 @ self.LU1.T
        self.V1 = self.LV1 @ self.LV1.T
        self.U2 = self.LU2 @ self.LU2.T
        self.V2 = self.LV2 @ self.LV2.T

        # Cholesky Kronecker decompositions
        self.cholkroncov1 = (self.LU1, self.LV1)
        self.cholkroncov2 = (self.LU2, self.LV2)

        # Full covariance matrices
        self.kron_covmat1 = torch.kron(self.U1, self.V1)
        self.kron_covmat2 = torch.kron(self.U2, self.V2)

        # Other inputs
        self.I = torch.eye(d, dtype=dtype)
        self.v1 = torch.randn(d, dtype=dtype)
        self.v2 = torch.randn(d, dtype=dtype)
        self.s2 = torch.tensor(2.0, dtype=dtype)
        self.a1 = torch.randn(d1, dtype=dtype).square()
        self.b1 = torch.randn(d2, dtype=dtype).square()

        self.rtol = 1e-6
        self.atol = 1e-6


    # auto_kl_*_diagvar
    def test_auto_kl_covmat_diagvar(self):
        optimal_diagvar2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvar')
        result_manual = kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvar', optimal_diagvar2, verbose=True)
        result_auto = auto_kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvar')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    def test_auto_kl_cholkroncov_diagvar(self):
        optimal_diagvar2 = optimal_covariance(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'diagvar')
        result_manual = kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'diagvar', optimal_diagvar2, verbose=True)
        result_auto = auto_kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'diagvar')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    # auto_kl_*_diagprec
    def test_auto_kl_covmat_diagprec(self):
        optimal_diagprec2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagprec')
        result_manual = kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagprec', optimal_diagprec2)
        result_auto = auto_kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagprec')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    # auto_kl_*_scalarvar
    def test_auto_kl_covmat_scalarvar(self):
        optimal_scalarvar2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarvar')
        result_manual = kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarvar', optimal_scalarvar2)
        result_auto = auto_kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarvar')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    def test_auto_kl_cholkroncov_scalarvar(self):
        optimal_scalarvar2 = optimal_covariance(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarvar')
        result_manual = kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarvar', optimal_scalarvar2)
        result_auto = auto_kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarvar')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    # auto_kl_*_scalarprec
    def test_auto_kl_covmat_scalarprec(self):
        optimal_scalarprec2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarprec')
        result_manual = kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarprec', optimal_scalarprec2)
        result_auto = auto_kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarprec')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    def test_auto_kl_cholkroncov_scalarprec(self):
        optimal_scalarprec2 = optimal_covariance(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarprec')
        result_manual = kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarprec', optimal_scalarprec2)
        result_auto = auto_kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarprec')

        self.assertTrue(torch.allclose(result_manual, result_auto))



# Run the tests
if __name__ == '__main__':
    unittest.main(argv=[''], exit=False)




""" Functions that compute trace term of the KL divergence for different covariance types. """

import torch

# trace_covmat_*

def trace_covmat_covmat(covmat1, covmat2):
    """ Computes trace term: Tr(M2^{-1} M1) / 2 """
    return 0.5 * torch.trace(torch.inverse(covmat2) @ covmat1)
    
def trace_covmat_precmat(covmat1, precmat2):
    """ Computes trace term: Tr(M2 M1) / 2 """
    return 0.5 * torch.trace(precmat2 @ covmat1)
    
def trace_covmat_diagvar(covmat1, diagvar2):
    """ Computes trace term: Tr(diag(v2)^{-1} M1) / 2 """
    return 0.5 * torch.sum(covmat1.diag() / diagvar2)

def trace_covmat_diagprec(covmat1, diagprec2):
    """ Computes trace term: Tr(diag(v2) M1) / 2 """
    return 0.5 * torch.sum(covmat1.diag() * diagprec2)

def trace_covmat_scalarvar(covmat1, scalarvar2):
    """ Computes trace term: Tr(scalarvar2^{-1} covmat1) / 2 """
    return 0.5 * torch.trace(covmat1) / scalarvar2.abs().squeeze()

def trace_covmat_scalarprec(covmat1, scalarprec2):
    """ Computes trace term: Tr(scalarprec2 I covmat1) / 2 """
    return 0.5 * torch.trace(covmat1) * scalarprec2.abs().squeeze()

def trace_covmat_identity(covmat1, _):
    """ Computes trace term: Tr(I covmat1)/2 """
    return 0.5 * torch.trace(covmat1)

def trace_covmat_cholcov(covmat1, cholcov2):
    """ Computes trace term: Tr((U2 kron V2)^{-1} M1)/2 """
    U2, V2 = cholcov2
    U2_inv = torch.inverse(U2)
    V2_inv = torch.inverse(V2)
    return 0.5 * torch.trace(torch.kron(U2_inv, V2_inv) @ covmat1)
    
def trace_covmat_cholprec(covmat1, cholprec2):
    """ Computes trace term: Tr((U2 ⊗ V2) M1)/2 """
    U2, V2 = cholprec2
    return 0.5 * torch.trace(torch.kron(U2, V2) @ covmat1)

def trace_covmat_kroncov(covmat1, kroncov2):
    """ Computes trace term: Tr((U2 kron V2)^{-1} M1) / 2 """
    U2, V2 = kroncov2
    UV2 = torch.kron(U2, V2)
    return 0.5 * torch.trace(torch.linalg.solve(UV2, covmat1))
    
def trace_covmat_kronprec(covmat1, kronprec2):
    """ Computes trace term: Tr((U2 kron V2) M1) / 2 """
    U2, V2 = kronprec2
    UV2 = torch.kron(U2, V2)
    return 0.5 * torch.trace(UV2 @ covmat1)

def trace_covmat_cholkroncov(covmat1, cholkroncov2):
    """ Computes trace term: Tr((LU2 LU2^T kron LV2 LV2^T)^{-1} M1) / 2 """
    LU2, LV2 = cholkroncov2
    U2 = LU2 @ LU2.T
    V2 = LV2 @ LV2.T
    U2_inv = torch.inverse(U2)
    V2_inv = torch.inverse(V2)
    
    return 0.5 * torch.trace(torch.kron(U2_inv, V2_inv) @ covmat1)
    
def trace_covmat_cholkronprec(covmat1, cholkronprec2):
    """ Computes trace term: Tr((LU2 LU2^T kron LV2 LV2^T) M1) / 2 """
    LU2, LV2 = cholkronprec2
    U2 = LU2 @ LU2.T
    V2 = LV2 @ LV2.T
    
    return 0.5 * torch.trace(torch.kron(U2, V2) @ covmat1)

def trace_covmat_diagvarrowcol(covmat1, diagvarrowcol2):
    """ Computes Tr(diag(vec(ba^T))^{-1} M1)/2 """
    diagvarrow, diagvarcol = diagvarrowcol2
    rowcol = torch.outer(diagvarrow, diagvarcol).view(-1)
    return 0.5 * torch.sum(covmat1.diag() / rowcol)

def trace_covmat_diagvarrow(covmat1, diagvarrow2):
    """ Computes Tr(diag(vec(1a^T))^{-1} M1)/2 """
    d1 = len(diagvarrow2)
    d2 = covmat1.size(0) // d1
    rowcol = diagvarrow2.view(-1, 1).repeat(1, d2).view(-1)
    return 0.5 * torch.sum(covmat1.diag() / rowcol)

def trace_covmat_diagvarcol(covmat1, diagvarcol2):
    """ Computes Tr(diag(vec(b1^T))^{-1} M1)/2 """
    d2 = len(diagvarcol2)
    d1 = covmat1.size(0) // d2
    rowcol = diagvarcol2.view(1, -1).repeat(d1, 1).view(-1)
    return 0.5 * torch.sum(covmat1.diag() / rowcol)


# trace_cholkroncov_*

def trace_cholkroncov_covmat(cholkroncov1, covmat2):
    """ Computes trace term: Tr(M2^{-1} (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    return 0.5 * torch.trace(torch.inverse(covmat2) @ torch.kron(U1, V1))

def trace_cholkroncov_precmat(cholkroncov1, precmat2):
    """ Computes trace term: Tr(M2 (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    return 0.5 * torch.trace(precmat2 @ torch.kron(U1, V1))

def trace_cholkroncov_diagvar(cholkroncov1, diagvar2):
    """ Computes trace term: Tr(diag(v2)^{-1} (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    U1_diag = LU1.square().sum(1)
    V1_diag = LV1.square().sum(1)
    return 0.5 * torch.sum(torch.outer(U1_diag, V1_diag).view(-1) / diagvar2)

def trace_cholkroncov_diagprec(cholkroncov1, diagprec2):
    """ Computes trace term: Tr(diag(v2) (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    U1_diag = LU1.square().sum(1)
    V1_diag = LV1.square().sum(1)
    return 0.5 * torch.sum(torch.outer(U1_diag, V1_diag).view(-1)  * diagprec2)

def trace_cholkroncov_scalarvar(cholkroncov1, scalarvar2):
    """ Computes trace term: Tr(scalarvar2^{-1} (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    return 0.5 * torch.trace(LU1 @ LU1.T) * torch.trace(LV1 @ LV1.T) / scalarvar2

def trace_cholkroncov_scalarprec(cholkroncov1, scalarprec2):
    """ Computes trace term: Tr(scalarprec2 I (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    return 0.5 * torch.trace(LU1 @ LU1.T) * torch.trace(LV1 @ LV1.T) * scalarprec2

def trace_cholkroncov_identity(cholkroncov1, _):
    """ Computes trace term: Tr(I (LU1 LU1^T kron LV1 LV1^T))/2 """
    LU1, LV1 = cholkroncov1
    return 0.5 * torch.trace(LU1 @ LU1.T) * torch.trace(LV1 @ LV1.T)

def trace_cholkroncov_cholcov(cholkroncov1, cholcov2):
    """ Computes trace term: Tr((U2 kron V2)^{-1} (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    U2, V2 = cholcov2
    U2_inv = torch.inverse(U2)
    V2_inv = torch.inverse(V2)
    return 0.5 * torch.trace(LU1.T @ U2_inv @ LU1) * torch.trace(LV1.T @ V2_inv @ LV1)

def trace_cholkroncov_cholprec(cholkroncov1, cholprec2):
    """ Computes trace term: Tr((U2 ⊗ V2) (LU1 LU1^T kron LV1 LV1^T))/2 """
    LU1, LV1 = cholkroncov1
    U2, V2 = cholprec2
    return 0.5 * torch.trace(LU1.T @ U2 @ LU1) * torch.trace(LV1.T @ V2 @ LV1)

def trace_cholkroncov_kroncov(cholkroncov1, kroncov2):
    """ Computes trace term: Tr((U2 kron V2)^{-1} (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    U2, V2 = kroncov2
    U2_inv = torch.inverse(U2)
    V2_inv = torch.inverse(V2)
    return 0.5 * torch.trace(LU1.T @ U2_inv @ LU1) * torch.trace(LV1.T @ V2_inv @ LV1)

def trace_cholkroncov_kronprec(cholkroncov1, kronprec2):
    """ Computes trace term: Tr((U2 kron V2) (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    U2, V2 = kronprec2
    return 0.5 * torch.trace(LU1.T @ U2 @ LU1) * torch.trace(LV1.T @ V2 @ LV1)

def trace_cholkroncov_cholkroncov(cholkroncov1, cholkroncov2):
    """ Computes trace term: Tr((LU2 LU2^T kron LV2 LV2^T)^{-1} (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    LU2, LV2 = cholkroncov2
    LU2_inv = torch.inverse(LU2)
    LV2_inv = torch.inverse(LV2)

    LU1TLU2_invT = LU1.T @ LU2_inv.T
    trace_LU1TLU2_invT = LU1TLU2_invT.square().sum()
    LV1TLV2_invT = LV1.T @ LV2_inv.T
    trace_LV1TLV2_invT = LV1TLV2_invT.square().sum()

    return 0.5 * trace_LU1TLU2_invT * trace_LV1TLV2_invT

def trace_cholkroncov_cholkronprec(cholkroncov1, cholkronprec2):
    """ Computes trace term: Tr((LU2 LU2^T kron LV2 LV2^T) (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    LU2, LV2 = cholkronprec2
    LU1TLU2 = LU1.T @ LU2
    trace_U2U1 = LU1TLU2.square().sum()
    LV1TLV2 = LV1.T @ LV2
    trace_V2V1 = LV1TLV2.square().sum()

    return 0.5 * trace_U2U1 * trace_V2V1

